use strict;
use warnings;

# this test was generated with Dist::Zilla::Plugin::Test::NoTabs 0.15

use Test::More 0.88;
use Test::NoTabs;

my @files = (
    'lib/Iterator/Flex.pm',
    'lib/Iterator/Flex/Array.pm',
    'lib/Iterator/Flex/ArrayLike.pm',
    'lib/Iterator/Flex/Base.pm',
    'lib/Iterator/Flex/Buffer.pm',
    'lib/Iterator/Flex/Cache.pm',
    'lib/Iterator/Flex/Cat.pm',
    'lib/Iterator/Flex/Chunk.pm',
    'lib/Iterator/Flex/Common.pm',
    'lib/Iterator/Flex/Cycle.pm',
    'lib/Iterator/Flex/Factory.pm',
    'lib/Iterator/Flex/Failure.pm',
    'lib/Iterator/Flex/Flatten.pm',
    'lib/Iterator/Flex/Freeze.pm',
    'lib/Iterator/Flex/Gather.pm',
    'lib/Iterator/Flex/Gather/Constants.pm',
    'lib/Iterator/Flex/Grep.pm',
    'lib/Iterator/Flex/Manual.pod',
    'lib/Iterator/Flex/Manual/Alien.pod',
    'lib/Iterator/Flex/Manual/Authoring.pod',
    'lib/Iterator/Flex/Manual/Basics.pod',
    'lib/Iterator/Flex/Manual/Caveats.pod',
    'lib/Iterator/Flex/Manual/Glossary.pod',
    'lib/Iterator/Flex/Manual/Internals.pod',
    'lib/Iterator/Flex/Manual/Overview.pod',
    'lib/Iterator/Flex/Manual/PriorArt.pod',
    'lib/Iterator/Flex/Manual/Serialization.pod',
    'lib/Iterator/Flex/Manual/Using.pod',
    'lib/Iterator/Flex/Map.pm',
    'lib/Iterator/Flex/Method.pm',
    'lib/Iterator/Flex/Permute.pm',
    'lib/Iterator/Flex/Product.pm',
    'lib/Iterator/Flex/Role/Current/Closure.pm',
    'lib/Iterator/Flex/Role/Current/Method.pm',
    'lib/Iterator/Flex/Role/Error/Throw.pm',
    'lib/Iterator/Flex/Role/Exhaustion/ImportedReturn.pm',
    'lib/Iterator/Flex/Role/Exhaustion/ImportedThrow.pm',
    'lib/Iterator/Flex/Role/Exhaustion/PassthroughThrow.pm',
    'lib/Iterator/Flex/Role/Exhaustion/Return.pm',
    'lib/Iterator/Flex/Role/Exhaustion/Throw.pm',
    'lib/Iterator/Flex/Role/Freeze.pm',
    'lib/Iterator/Flex/Role/Next/ClosedSelf.pm',
    'lib/Iterator/Flex/Role/Next/Closure.pm',
    'lib/Iterator/Flex/Role/Prev/Closure.pm',
    'lib/Iterator/Flex/Role/Prev/Method.pm',
    'lib/Iterator/Flex/Role/Reset/Closure.pm',
    'lib/Iterator/Flex/Role/Reset/Method.pm',
    'lib/Iterator/Flex/Role/Rewind/Closure.pm',
    'lib/Iterator/Flex/Role/Rewind/Method.pm',
    'lib/Iterator/Flex/Role/State.pm',
    'lib/Iterator/Flex/Role/State/Closure.pm',
    'lib/Iterator/Flex/Role/State/Registry.pm',
    'lib/Iterator/Flex/Role/Wrap/Return.pm',
    'lib/Iterator/Flex/Role/Wrap/Self.pm',
    'lib/Iterator/Flex/Role/Wrap/Throw.pm',
    'lib/Iterator/Flex/Sequence.pm',
    'lib/Iterator/Flex/Stack.pm',
    'lib/Iterator/Flex/Take.pm',
    'lib/Iterator/Flex/Utils.pm',
    'lib/Iterator/Flex/Zip.pm',
    't/00-compile.t',
    't/00-report-prereqs.dd',
    't/00-report-prereqs.t',
    't/ArrayLike.t',
    't/Base/added_methods.t',
    't/Base/error.t',
    't/Base/exhaustion.t',
    't/Base/input_exhaustion.t',
    't/Base/may.t',
    't/Base/methods.t',
    't/Base/namespaces.t',
    't/Common/iarray.t',
    't/Common/ibuffer.t',
    't/Common/icache.t',
    't/Common/icat.t',
    't/Common/ichain.t',
    't/Common/ichunk.t',
    't/Common/icycle.t',
    't/Common/iflatten.t',
    't/Common/ifreeze.t',
    't/Common/igather.t',
    't/Common/igrep.t',
    't/Common/imap.t',
    't/Common/iproduct.t',
    't/Common/iseq.t',
    't/Common/istack.t',
    't/Common/itake.t',
    't/Common/iterator.t',
    't/Common/izip.t',
    't/Factory/construct_from_attrs.t',
    't/Factory/construct_from_iterable.t',
    't/Permute.t',
    't/Role/Utils.t',
    't/lib/MyTest/Tests/May/Iter.pm',
    't/lib/MyTest/Utils.pm',
    't/multi_stage.t',
    't/regressions/next-active.t'
);

notabs_ok($_) foreach @files;
done_testing;
