use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007006
use Test::Spelling 0.17;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Alien
Array
ArrayLike
Astrophysical
Authoring
Base
Basics
Buffer
Cache
Cat
Caveats
Chunk
ClosedSelf
Closure
Common
Constants
Current
Cycle
Diab
Error
Exhaustion
Factory
Failure
Flatten
Flex
Freeze
Gather
Glossary
Grep
ImportedReturn
ImportedThrow
Internals
Iterator
Jerius
Manual
Map
Method
Next
Observatory
Overview
PassthroughThrow
Permute
Prev
PriorArt
Product
Registry
Reset
Return
Rewind
Role
Self
Sequence
Serialization
Smithsonian
Stack
State
Take
Throw
Using
Utils
Wrap
Zip
djerius
iarray
icache
icycle
ifreeze
igrep
imap
iproduct
iseq
iter
iterable
iterables
lib
queryable
